import tfkit
from transformers import pipeline, pipelines, BertTokenizer
import inquirer
from .parser import *


class Model:

    def __init__(self, model_path, model_task=None, enable_arg_panel=False):
        self.model = None
        if nlp2.is_file_exist(model_path) or "tfkit_" in model_path:  # tfkit models
            self.lib = 'tfkit'
            self.model, self.predict_parameter, self.model_task = self.load_tfkit_model(model_path, model_task,
                                                                                        enable_arg_panel)
        else:  # huggingface's transfromers model - local model saved in dir, online model name without tfkit tag
            self.lib = 'hf'
            self.model, self.predict_parameter, self.model_task = self.load_huggingface_model(model_path, model_task,
                                                                                              enable_arg_panel)

    def load_huggingface_model(self, model_path, model_task=None, enable_arg_panel=False):
        supported_type = list(pipelines.SUPPORTED_TASKS.keys())
        if model_task is None or model_task not in supported_type:
            inquirer_res = inquirer.prompt(
                [inquirer.List('model_task', message="Select model task", choices=supported_type)])
            model_task = inquirer_res['model_task']

        tok_conf = BertTokenizer.from_pretrained(model_path) if 'voidful/albert_chinese' in model_path else model_path
        nlp = pipeline(model_task, model=model_path,
                       tokenizer=tok_conf)
        predict_parameter = {}
        return nlp, predict_parameter, model_task

    def load_tfkit_model(self, model_path, model_task=None, enable_arg_panel=False):
        model = tfkit.load_model(model_path, model_task)
        predict_parameter = tfkit.load_predict_parameter(model, enable_arg_panel=enable_arg_panel)
        model_task = model.__class__.__name__
        return model, predict_parameter, model_task

    def predict(self, argument={}, enable_input_panel=False):
        parser = Parser(self.model_task)
        input_argument = parser.inputParser(argument, enable_input_panel)

        if "wrong" in input_argument:
            return {
                'result': 'incorrect input:' + str(input_argument['wrong']) +
                          ', Should be:' + str(input_argument["all"]),
                'result_info': input_argument
            }
        else:
            if self.lib == 'hf':
                predict_func = self.model
            else:
                predict_func = self.model.predict

            if isinstance(input_argument, str):
                return parser.outputParser(predict_func(input_argument))
            else:
                self.predict_parameter.update(input_argument)
                return parser.outputParser(predict_func(**self.predict_parameter))



import json
from .util import NumpyEncoder


class Cli:

    def start(self, models):
        while True:
            for path, model in models.items():
                result_dict = model.predict(enable_input_panel=True)
                print(json.dumps(result_dict['result'], ensure_ascii=False, cls=NumpyEncoder, indent=4, sort_keys=True))



import logging
from flask import Flask, request, Response
from flask_cors import CORS
from gevent.pywsgi import WSGIServer
import json
from .util import NumpyEncoder

logger = logging.getLogger(__name__)


class Server:
    class ServerError(Exception):
        status_code = 400

        def __init__(self, message, status_code=None, payload=None):
            Exception.__init__(self)
            self.message = message
            if status_code is not None:
                self.status_code = status_code
            self.payload = payload

        def to_dict(self):
            error_dict = dict(self.payload or ())
            error_dict['message'] = self.message
            return error_dict

    def make_app(self, models) -> Flask:
        app = Flask(__name__)  # pylint: disable=invalid-name

        @app.route('/api/<path>', methods=['GET', 'POST'])
        def predict(path) -> Response:
            if path in models:
                result_dict = models[path].predict(dict(request.values), enable_input_panel=False)
                return json.dumps({'result': result_dict['result']}, ensure_ascii=False, cls=NumpyEncoder,
                                  indent=4,
                                  sort_keys=True)
            else:
                raise self.ServerError("parameter not found", 404)

        return app

    def start(self, models, port):
        print("hosting api in path: /api/+", list(models.keys()))
        app = self.make_app(models)
        CORS(app)

        http_server = WSGIServer(('0.0.0.0', port), app)
        print(f"Model loaded, serving demo on port {port}")
        http_server.serve_forever()



import argparse
from collections import defaultdict
import json

from .cli import Cli
from .model import Model
from .server import Server


def main():
    """
    loading argument and starting an model interface
    """
    global model_dict
    model_dict = defaultdict(dict)
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--json', type=str)
    group.add_argument('--model', type=str)
    parser.add_argument('--enable_arg_panel', action='store_true', help='enable panel to input argument')
    parser.add_argument('--task', type=str)
    # interface
    parser.add_argument('--api_path', type=str, default="model", help='api path to serve the demo')
    parser.add_argument('--api_port', type=int, default=3000, help='port to serve the demo on')
    parser.add_argument('--cli', action='store_true', help='commandline mode')

    args = parser.parse_args()

    if args.model:
        model_dict[args.api_path] = Model(args.model, args.task, args.enable_arg_panel)
    else:
        with open(args.json, 'r', encoding='utf8') as reader:
            model_dict = json.loads(reader.read())

        for path, d in model_dict.items():
            task = model_dict[path]['task'] if 'task' in model_dict[path] else None
            model = Model(model_dict[path]['model'], task, args.enable_arg_panel)
            model_dict[path] = model

    if args.cli:
        cli = Cli()
        cli.start(model_dict)
    else:
        server = Server()
        server.start(model_dict, args.api_port)


if __name__ == "__main__":
    main()



from collections import OrderedDict
import nlp2


class Parser:
    def __init__(self, model_type):
        self.input_parser = self.INPUT_PARSER_MAPPING[model_type]
        self.output_parser = self.OUTPUT_PARSER_MAPPING[model_type]

    def inputParser(self, argument={}, enable_arg_panel=False):
        argument = nlp2.function_argument_panel(self.input_parser, inputted_arg=argument,
                                                disable_input_panel=(not enable_arg_panel))
        # error input check
        wrong = nlp2.function_check_wrong_arg(self.input_parser, list(argument.keys()))
        if len(wrong) > 0:
            return {"wrong": wrong, "all": nlp2.function_get_all_arg(self.input_parser)}
        else:
            return self.input_parser(self, **argument)

    def outputParser(self, *outputs):
        return self.output_parser(self, outputs)

    # Specific Model Input Parser
    def inputTfkitGeneralParser(self, input=""):
        # sep = tfkit.utility.tok_sep(self.model.tokenizer)
        # sep = sep.replace('[', '\[').replace(']', '\]').replace('<', '\<').replace('>', '\>')
        # regex = r"\[Question\]|[0-9]+|[a-zA-Z]+\'*[a-z]*|[\w\W]" + "|" + sep
        # input = " ".join(re.findall(regex, input, re.UNICODE))
        return {'input': input}

    def inputHFGeneralParser(self, input=""):
        return input

    def inputTfkitQAParser(self, passage="", question=""):
        return {'input': passage + "[SEP]" + question}

    def inputHFQAParser(self, context="", question=""):
        return {'context': context, 'question': question}

    # Specific Model Output Parser
    def outputHFGeneralParser(self, result_list):
        answer_list = []
        for r in result_list:
            answer_list.append(r)
        return {'result': answer_list}

    def outputTfkitGeneralParser(self, output):
        (result_list, result_dict) = output[0]
        return {
            'result': result_list,
            'result_info': result_dict
        }

    def outputHFQAParser(self, result_list):
        answer_list = []
        for r in result_list:
            answer_list.append(r['answer'])
        return {'result': answer_list, 'result_info': result_list}

    OUTPUT_PARSER_MAPPING = OrderedDict(
        [
            # Huggingface's model
            ("feature-extraction", outputHFGeneralParser,),
            ("sentiment-analysis", outputHFGeneralParser,),
            ("ner", outputHFGeneralParser,),
            ("question-answering", outputHFQAParser,),
            ("fill-mask", outputHFGeneralParser,),
            ("summarization", outputHFGeneralParser,),
            ("translation_en_to_fr", outputHFGeneralParser,),
            ("translation_en_to_de", outputHFGeneralParser,),
            ("translation_en_to_ro", outputHFGeneralParser,),
            ("text-generation", outputHFGeneralParser,),
            # TFKIT model
            ("Tagger", outputTfkitGeneralParser,),
            ("QA", outputTfkitGeneralParser,),
            ("Twice", outputTfkitGeneralParser,),
            ("OneByOne", outputTfkitGeneralParser,),
            ("Once", outputTfkitGeneralParser,),
            ("BiDiOneByOne", outputTfkitGeneralParser,),
            ("MtClassifier", outputTfkitGeneralParser,),
        ]
    )

    INPUT_PARSER_MAPPING = OrderedDict(
        [
            # Huggingface's model
            ("feature-extraction", inputHFGeneralParser,),
            ("sentiment-analysis", inputHFGeneralParser,),
            ("ner", inputHFGeneralParser,),
            ("question-answering", inputHFQAParser,),
            ("fill-mask", inputHFGeneralParser,),
            ("summarization", inputHFGeneralParser,),
            ("translation_en_to_fr", inputHFGeneralParser,),
            ("translation_en_to_de", inputHFGeneralParser,),
            ("translation_en_to_ro", inputHFGeneralParser,),
            ("text-generation", inputHFGeneralParser,),
            # TFKIT model
            ("Tagger", inputTfkitGeneralParser,),
            ("QA", inputTfkitQAParser,),
            ("Twice", inputTfkitGeneralParser,),
            ("OneByOne", inputTfkitGeneralParser,),
            ("Once", inputTfkitGeneralParser,),
            ("BiDiOneByOne", inputTfkitGeneralParser,),
            ("MtClassifier", inputTfkitGeneralParser,),
        ]
    )



from setuptools import setup, find_packages

setup(
    name='nlp2go',
    version='0.1.0',
    description='hosting nlp models for demo purpose',
    url='https://github.com/voidful/nlp2go',
    author='Voidful',
    author_email='voidful.stack@gmail.com',
    long_description=open("README.md", encoding="utf8").read(),
    long_description_content_type="text/markdown",
    keywords='nlp tfkit classification generation tagging deep learning machine reading',
    packages=find_packages(),
    install_requires=[
        "tfkit>=0.3.10",
        "flask",
        "flask-cors",
        "gevent",
        "nlp2>=1.6.5"
    ],
    entry_points={
        'console_scripts': ['nlp2go=nlp2go.main:main', 'nlp2go-preload=nlp2go.preload:main']
    },
    zip_safe=False,
)



import json
import numpy as np


class NumpyEncoder(json.JSONEncoder):
    """ Special json encoder for numpy types """

    def default(self, obj):
        if isinstance(obj, (np.int_, np.intc, np.intp, np.int8,
                            np.int16, np.int32, np.int64, np.uint8,
                            np.uint16, np.uint32, np.uint64)):
            return int(obj)
        elif isinstance(obj, (np.float_, np.float16, np.float32,
                              np.float64)):
            return float(obj)
        elif isinstance(obj, (np.ndarray,)):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)



import argparse
from collections import defaultdict

import json
from .model import Model


def main():
    global model_dict
    model_dict = defaultdict(dict)
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--json', type=str)
    group.add_argument('--model', type=str)
    args = parser.parse_args()

    if args.model:
        Model(args.model)
    else:
        with open(args.json, 'r', encoding='utf8') as reader:
            model_dict = json.loads(reader.read())
        for k, v in model_dict.items():
            Model(model_dict[k]['model'])

    print("==================")
    print("finish pre loading")


if __name__ == "__main__":
    main()
